import 'package:riverflow/src/extractors/extractor.dart';
import 'package:riverflow/src/extractors/extractor_types.dart';
import 'package:riverflow/src/util/html_utils.dart';
import 'package:html/dom.dart' as dom;

class HtmlExtractor extends Extractor {
  final String selector;
  final List<String> collectors;
  final int maxCount; 

  HtmlExtractor({
    this.selector,
    this.collectors,
    this.maxCount = 0
  }) : super(ExtractorTypes.HTML);

  @override
  List extract(dynamic input) {
    return selectMatchingElements(HtmlUtils.formatInputElement(input))
        .map((element) => HtmlUtils.applyCollectors(element, collectors))
        .where((element) => element != null)
        .toList();
  }

  List<dom.Element> selectMatchingElements(dom.Element element) {
    if (selector != null) {
      return (maxCount>1)? element.querySelectorAll(selector): [element.querySelector(selector)].where((element) => element!=null);
    } else {
      return [element];
    }
  }

  factory HtmlExtractor.fromJson(Map<String, dynamic> json) {
    return HtmlExtractor(
      selector: json['selector'],
      collectors: json['collectors'].map<String>((e) => e as String).toList(),
      maxCount: json.containsKey('max_count')?(json['max_count'] as int): 0,
    );
  }

  @override
  Map<String, dynamic> toJson() {
    final json = super.toJson();
    json['selector'] = selector;
    json['collectors'] = collectors;
        json['max_count'] = maxCount??0;
    return json;
  }
}

class HtmlIncludeExtractor extends Extractor {
  final String selector;
  final int maxCount; 

  HtmlIncludeExtractor({
    this.selector,
    this.maxCount = 0,
  }) : super(ExtractorTypes.HTML_INCLUDE);

  @override
  List extract(dynamic input) {
    return selectMatchingElements(HtmlUtils.formatInputElement(input))
        .where((element) => element != null)
        .toList();
  }

  List<dom.Element> selectMatchingElements(dom.Element element) {
    if (selector != null) {
      return (maxCount > 1)? element.querySelectorAll(selector): [element.querySelector(selector)].where((element) => element!=null);
    } else {
      return [element];
    }
  }

  factory HtmlIncludeExtractor.fromJson(Map<String, dynamic> json) {
    return HtmlIncludeExtractor(
      selector: json['selector']?.toString(),
      maxCount: json.containsKey('max_count')?(json['max_count'] as int): 0,
    );
  }

  @override
  Map<String, dynamic> toJson() {
    final json = super.toJson();
    json['selector'] = selector;
    json['max_count'] = maxCount??0;
    return json;
  }
}

class HtmlExcludeExtractor extends Extractor {
  final String selector;
  final int maxCount; 

  HtmlExcludeExtractor({
    this.selector,
    this.maxCount = 0,
  }) : super(ExtractorTypes.HTML_EXCLUDE);

  @override
  List extract(dynamic input) {
    return selectMatchingElements(HtmlUtils.formatInputElement(input))
        .where((element) => element != null)
        .toList();
  }

  List<dom.Element> selectMatchingElements(dom.Element element) {
    if (selector != null) {
      ((maxCount >1)? element.querySelectorAll(selector): [element.querySelector(selector)].where((element) => element!=null)).forEach((element) {
        element.remove();
      });
    }
    return [element];
  }

  factory HtmlExcludeExtractor.fromJson(Map<String, dynamic> json) {
    return HtmlExcludeExtractor(
      selector: json['selector'],
      maxCount: json.containsKey('max_count')?(json['max_count'] as int): 0,
    );
  }

  @override
  Map<String, dynamic> toJson() {
    final json = super.toJson();
    json['selector'] = selector;
    json['max_count'] = maxCount??0;
    return json;
  }
}

class HtmlAttributeEditor extends Extractor {
  final String selector;
  final Map<String, String> attributes;
  final int maxCount; 

  HtmlAttributeEditor({
    this.selector,
    this.attributes,
    this.maxCount = 0,
  }) : super(ExtractorTypes.HTML_CHANGE_ATTRIBUTE);

  @override
  List extract(dynamic input) {
    selectMatchingElements(HtmlUtils.formatInputElement(input)).forEach((element) {
      if (element.attributes != null) {
        final descAttrs = element.attributes?.keys
                ?.where((attr) => attr != null)
                ?.map((attr) => attr as String)
                ?.where((attr) => attributes.containsKey(attr))
                ?.where((attr) => attributes[attr]?.trim()?.isNotEmpty ?? false)
                ?.where((attr) => element.attributes[attr]?.trim()?.isNotEmpty ?? false)
                ?.toList() ??
            [];
        descAttrs.forEach((attr) {
          final srcAttr = attributes[attr];
          element.attributes[srcAttr] = element.attributes[attr];
        });
      }
    });
    return [input];
  }

  List<dom.Element> selectMatchingElements(dom.Element element) {
    if (selector != null) {
      return (maxCount > 1)? element.querySelectorAll(selector) : [element.querySelector(selector)].where((element) => element!=null);
    } else {
      return [element];
    }
  }

  factory HtmlAttributeEditor.fromJson(Map<String, dynamic> json) {
    return HtmlAttributeEditor(
      selector: json['selector']?.toString(),
      attributes:
          (json['attributes'] as Map).map((key, value) => MapEntry<String, String>(key as String, value as String)),
        maxCount: json.containsKey('max_count')?(json['max_count'] as int): 0,
    );
  }

  @override
  Map<String, dynamic> toJson() {
    final json = super.toJson();
    json['selector'] = selector;
    json['attributes'] = attributes;
        json['max_count'] = maxCount??0;
    return json;
  }
}

class HtmlRenameExtractor extends Extractor {
  final String selector;
  final String renameTo;
  final int maxCount; 

  HtmlRenameExtractor({
    this.selector,
    this.renameTo,
    this.maxCount = 0,
  }) : super(ExtractorTypes.HTML_TAG_RENAME);

  @override
  List extract(dynamic input) {
    final root = HtmlUtils.formatInputElement(input);
    selectMatchingElements(root).forEach((element) {
      element.replaceWith(HtmlUtils.cloneWithName(element, renameTo));
    });
    return [root];
  }

  List<dom.Element> selectMatchingElements(dom.Element element) {
    if (selector != null) {
      return (maxCount > 1)? element.querySelectorAll(selector): [element.querySelector(selector)].where((element) => element!=null);
    } else {
      return [element];
    }
  }

  factory HtmlRenameExtractor.fromJson(Map<String, dynamic> json) {
    return HtmlRenameExtractor(
      selector: json['selector']?.toString(),
      renameTo: json['rename_to']?.toString(),
    maxCount: json.containsKey('max_count')?(json['max_count'] as int): 0,
    );
  }

  @override
  Map<String, dynamic> toJson() {
    final json = super.toJson();
    json['selector'] = selector;
    json['rename_to'] = renameTo;
    json['max_count'] = maxCount??0;
    return json;
  }
}

class HtmlDecoderExtractor extends Extractor {
  HtmlDecoderExtractor() : super(ExtractorTypes.HTML_DECODE);

  @override
  List extract(dynamic input) {
    final text = HtmlUtils.formatInputElement(input).text.trim();
    return [text];
  }

  factory HtmlDecoderExtractor.fromJson(Map<String, dynamic> json) {
    return HtmlDecoderExtractor();
  }
}

class HtmlTextExtractor extends Extractor {
  HtmlTextExtractor() : super(ExtractorTypes.HTML_TEXT);

  @override
  List extract(dynamic input) {
    final text = HtmlUtils.formatInputElement(input).text.trim();
    return [text];
  }

  factory HtmlTextExtractor.fromJson(Map<String, dynamic> json) {
    return HtmlTextExtractor();
  }
}
